@@ -2,19 +2,7 @@ import asyncio
 from collections import deque
 from datetime import datetime
 from io import BytesIO
-from typing import (
-    Any,
-    AsyncGenerator,
-    cast,
-    Deque,
-    Dict,
-    List,
-    Optional,
-    Set,
-    Tuple,
-    TYPE_CHECKING,
-    Union,
-)
+from typing import Any, AsyncGenerator, cast, Optional, TYPE_CHECKING, Union
 
 import magic
 from linkedin_messaging import URN
@@ -22,8 +10,10 @@ from linkedin_messaging.api_objects import (
     AttributedBody,
     Conversation,
     ConversationEvent,
+    Error,
     MessageAttachment,
     MessageCreate,
+    MiniProfile,
     ReactionSummary,
     RealTimeEventStreamEvent,
     ThirdPartyMedia,
@@ -94,14 +84,14 @@ MediaInfo = Union[FileInfo, VideoInfo, AudioInfo, ImageInfo]
 
 class Portal(DBPortal, BasePortal):
     invite_own_puppet_to_pm: bool = False
-    by_mxid: Dict[RoomID, "Portal"] = {}
-    by_li_thread_urn: Dict[Tuple[URN, Optional[URN]], "Portal"] = {}
+    by_mxid: dict[RoomID, "Portal"] = {}
+    by_li_thread_urn: dict[tuple[URN, Optional[URN]], "Portal"] = {}
     matrix: "MatrixHandler"
     config: Config
 
     backfill_lock: SimpleLock
-    _dedup: Deque[URN]
-    _send_locks: Dict[URN, asyncio.Lock]
+    _dedup: deque[URN]
+    _send_locks: dict[URN, asyncio.Lock]
     _noop_lock: FakeLock = FakeLock()
 
     def __init__(
@@ -114,8 +104,9 @@ class Portal(DBPortal, BasePortal):
         name: Optional[str] = None,
         photo_id: Optional[str] = None,
         avatar_url: Optional[ContentURI] = None,
+        topic: Optional[str] = None,
         encrypted: bool = False,
-    ) -> None:
+    ):
         super().__init__(
             li_thread_urn,
             li_receiver_urn,
@@ -126,6 +117,7 @@ class Portal(DBPortal, BasePortal):
             name,
             photo_id,
             avatar_url,
+            topic,
         )
         self.log = self.log.getChild(self.li_urn_log)
 
@@ -138,10 +130,10 @@ class Portal(DBPortal, BasePortal):
         self.backfill_lock = SimpleLock(
             "Waiting for backfilling to finish before handling %s", log=self.log
         )
-        self._backfill_leave: Optional[Set[IntentAPI]] = None
+        self._backfill_leave: Optional[set[IntentAPI]] = None
 
     @classmethod
-    def init_cls(cls, bridge: "LinkedInBridge") -> None:
+    def init_cls(cls, bridge: "LinkedInBridge"):
         BasePortal.bridge = bridge
         cls.az = bridge.az
         cls.config = bridge.config
@@ -155,7 +147,7 @@ class Portal(DBPortal, BasePortal):
 
     # region DB conversion
 
-    async def delete(self) -> None:
+    async def delete(self):
         if self.mxid:
             await DBMessage.delete_all_by_room(self.mxid)
             self.by_mxid.pop(self.mxid, None)
@@ -186,7 +178,7 @@ class Portal(DBPortal, BasePortal):
     # region Properties
 
     @property
-    def li_urn_full(self) -> Tuple[URN, Optional[URN]]:
+    def li_urn_full(self) -> tuple[URN, Optional[URN]]:
         return self.li_thread_urn, self.li_receiver_urn
 
     @property
@@ -211,7 +203,7 @@ class Portal(DBPortal, BasePortal):
 
     # region Database getters
 
-    async def postinit(self) -> None:
+    async def postinit(self):
         self.by_li_thread_urn[self.li_urn_full] = self
         if self.mxid:
             self.by_mxid[self.mxid] = self
@@ -296,7 +288,7 @@ class Portal(DBPortal, BasePortal):
     @classmethod
     async def all(cls) -> AsyncGenerator["Portal", None]:
         portals = await super().all()
-        for portal in cast(List[Portal], portals):
+        for portal in cast(list[Portal], portals):
             try:
                 yield cls.by_li_thread_urn[
                     (portal.li_thread_urn, portal.li_receiver_urn)
@@ -313,7 +305,7 @@ class Portal(DBPortal, BasePortal):
         self,
         source: Optional["u.User"] = None,
         conversation: Optional[Conversation] = None,
-    ) -> Conversation:
+    ):
         if not conversation:
             # shouldn't happen currently
             assert False, "update_info called without conversation"
@@ -325,24 +317,77 @@ class Portal(DBPortal, BasePortal):
             )
 
         changed = False
-
-        if not self.is_direct:
-            # TODO (#53)
-            # changed = any(
-            #     await asyncio.gather(
-            #         self._update_name(info.name),
-            #         self._update_photo(source, info.image),
-            #         loop=self.loop,
-            #     )
-            # )
-            pass
+        if self.is_direct:
+            if (
+                len(conversation.participants)
+                and (mm := conversation.participants[0].messaging_member)
+                and (mp := mm.mini_profile)
+            ):
+                changed = await self._update_topic(mp) or changed
+        else:
+            changed = await self._update_name(conversation.name) or changed
 
         changed = await self._update_participants(source, conversation) or changed
         if changed:
             await self.update_bridge_info()
             await self.save()
 
-        return conversation
+    async def _update_name(self, name: str) -> bool:
+        if not name:
+            self.log.warning("Got empty name in _update_name call")
+            return False
+        if self.name != name:
+            self.log.trace("Updating name %s -> %s", self.name, name)
+            self.name = name
+            if self.mxid and (self.encrypted or not self.is_direct):
+                await self.main_intent.set_room_name(self.mxid, self.name)
+            return True
+        return False
+
+    async def _update_photo_from_puppet(self, puppet: "p.Puppet") -> bool:
+        if self.photo_id == puppet.photo_id:
+            return False
+        self.photo_id = puppet.photo_id
+        if puppet.photo_mxc:
+            self.avatar_url = puppet.photo_mxc
+        elif self.photo_id:
+            profile = await self.main_intent.get_profile(puppet.default_mxid)
+            self.avatar_url = profile.avatar_url
+            puppet.photo_mxc = profile.avatar_url
+        else:
+            self.avatar_url = ContentURI("")
+        if self.mxid:
+            await self.main_intent.set_room_avatar(self.mxid, self.avatar_url)
+        return True
+
+    async def _update_topic(self, mini_profile: MiniProfile) -> bool:
+        if not self.config["bridge.set_topic_on_dms"]:
+            return False
+
+        topic_parts = [
+            part
+            for part in [
+                mini_profile.occupation,
+                (
+                    f"https://www.linkedin.com/in/{mini_profile.public_identifier}"
+                    if (
+                        mini_profile.public_identifier
+                        and mini_profile.public_identifier != "UNKNOWN"
+                    )
+                    else None
+                ),
+            ]
+            if part
+        ]
+        topic = " | ".join(topic_parts) if len(topic_parts) else None
+        if topic == self.topic:
+            return False
+        self.topic = topic
+
+        if self.mxid:
+            await self.main_intent.set_room_topic(self.mxid, self.topic or "")
+
+        return True
 
     async def update_bridge_info(self):
         if not self.mxid:
@@ -375,7 +420,16 @@ class Portal(DBPortal, BasePortal):
         changed = False
 
         for participant in conversation.participants if conversation else []:
-            participant_urn = participant.messaging_member.mini_profile.entity_urn
+            if (
+                not (mm := participant.messaging_member)
+                or not (mp := mm.mini_profile)
+                or not (entity_urn := mp.entity_urn)
+            ):
+                self.log.error(f"No entity_urn on participant! {participant}")
+                continue
+            participant_urn = entity_urn
+            if participant_urn == URN("UNKNOWN"):
+                participant_urn = conversation.entity_urn
             puppet = await p.Puppet.get_by_li_member_urn(participant_urn)
             await puppet.update_info(source, participant.messaging_member)
             if (
@@ -383,10 +437,8 @@ class Portal(DBPortal, BasePortal):
                 and self.li_other_user_urn == puppet.li_member_urn
                 and self.encrypted
             ):
-                pass
-                # TODO (#53)
-                # changed = await self._update_name(puppet.name) or changed
-                # changed = await self._update_photo_from_puppet(puppet) or changed
+                changed = await self._update_name(puppet.name) or changed
+                changed = await self._update_photo_from_puppet(puppet) or changed
 
             if self.mxid:
                 if puppet.li_member_urn != self.li_receiver_urn or puppet.is_real_user:
@@ -466,20 +518,29 @@ class Portal(DBPortal, BasePortal):
             if self.is_direct:
                 invites.append(self.az.bot_mxid)
 
-        info = await self.update_info(source, conversation)
+        await self.update_info(source, conversation)
+
+        if self.topic:
+            initial_state.append(
+                {
+                    "type": str(EventType.ROOM_TOPIC),
+                    "content": {"topic": self.topic},
+                }
+            )
+
         # if not info:
         #     self.log.debug(
         #         "update_info() didn't return info, cancelling room creation")
         #     return None
 
-        # if self.encrypted or not self.is_direct:
-        #     name = self.name
-        #     initial_state.append(
-        #         {
-        #             "type": str(EventType.ROOM_AVATAR),
-        #             "content": {"avatar_url": self.avatar_url},
-        #         }
-        #     )
+        if self.encrypted or not self.is_direct:
+            name = self.name
+            initial_state.append(
+                {
+                    "type": str(EventType.ROOM_AVATAR),
+                    "content": {"avatar_url": self.avatar_url},
+                }
+            )
 
         # We lock backfill lock here so any messages that come between the room being
         # created and the initial backfill finishing wouldn't be bridged before the
@@ -507,8 +568,18 @@ class Portal(DBPortal, BasePortal):
             self.by_mxid[self.mxid] = self
 
             if not self.is_direct:
-                await self._update_participants(source, info)
+                await self._update_participants(source, conversation)
             else:
+                if (
+                    (mm := conversation.participants[0].messaging_member)
+                    and (mp := mm.mini_profile)
+                    and (mp.entity_urn == URN("UNKNOWN"))
+                ):
+                    levels = await self.main_intent.get_power_levels(self.mxid)
+                    if levels.get_user_level(self.main_intent.mxid) == 100:
+                        levels.events_default = 50
+                        await self.main_intent.set_power_levels(self.mxid, levels)
+
                 puppet = await p.Puppet.get_by_custom_mxid(source.mxid)
                 if puppet:
                     try:
@@ -548,7 +619,7 @@ class Portal(DBPortal, BasePortal):
         return f"com.github.linkedin://linkedin/{self.li_thread_urn.id_str()}"
 
     @property
-    def bridge_info(self) -> Dict[str, Any]:
+    def bridge_info(self) -> dict[str, Any]:
         return {
             "bridgebot": self.az.bot_mxid,
             "creator": self.main_intent.mxid,
@@ -618,10 +689,11 @@ class Portal(DBPortal, BasePortal):
         self,
         source: "u.User",
         limit: Optional[int],
-        after_timestamp: Optional[datetime],  # TODO (54)
+        after_timestamp: Optional[datetime],
         conversation: Conversation,
     ):
         assert self.mxid
+        assert conversation.entity_urn
         assert source.client, f"No client found for {source.mxid}!"
         self.log.debug(f"Backfilling history through {source.mxid}")
         messages = conversation.events
@@ -647,12 +719,50 @@ class Portal(DBPortal, BasePortal):
             if len(elements) < 20:
                 break
 
+            if (
+                len(elements)
+                and elements[0].created_at
+                and after_timestamp
+                and (created_at := elements[0].created_at) <= after_timestamp
+            ):
+                self.log.debug(
+                    f"Stopping fetching messages at {created_at} as message is older "
+                    f"than newest bridged message ({created_at} < {after_timestamp})",
+                )
+                break
+
             before_timestamp = messages[0].created_at
 
+        if after_timestamp:
+            try:
+                slice_index = next(
+                    index
+                    for index, message in enumerate(messages)
+                    if message.created_at and message.created_at > after_timestamp
+                )
+                messages = messages[slice_index:]
+            except StopIteration:
+                messages = []
+
+        if limit and len(messages) > limit:
+            messages = messages[-limit:]
+
         self._backfill_leave = set()
         async with NotificationDisabler(self.mxid, source):
             for message in messages:
-                member_urn = message.from_.messaging_member.mini_profile.entity_urn
+                if (
+                    not (f := message.from_)
+                    or not (mm := f.messaging_member)
+                    or not (mp := mm.mini_profile)
+                    or not (entity_urn := mp.entity_urn)
+                ):
+                    self.log.error(
+                        "No entity_urn found on message mini_profile!", message
+                    )
+                    continue
+                member_urn = entity_urn
+                if member_urn == URN("UNKNOWN"):
+                    member_urn = conversation.entity_urn
                 puppet = await p.Puppet.get_by_li_member_urn(member_urn)
                 await self.handle_linkedin_message(source, puppet, message)
         for intent in self._backfill_leave:
@@ -664,13 +774,37 @@ class Portal(DBPortal, BasePortal):
 
     # region Matrix event handling
 
-    async def _send_delivery_receipt(self, event_id: EventID) -> None:
+    async def _send_delivery_receipt(self, event_id: EventID):
         if event_id and self.config["bridge.delivery_receipts"] and self.mxid:
             try:
                 await self.az.intent.mark_read(self.mxid, event_id)
             except Exception:
                 self.log.exception(f"Failed to send delivery receipt for {event_id}")
 
+    async def _send_bridge_error(self, msg: str, certain_failure: bool = False):
+        certainty = "was not" if certain_failure else "may not have been"
+        await self._send_message(
+            self.main_intent,
+            TextMessageEventContent(
+                msgtype=MessageType.NOTICE,
+                body=f"\u26a0 Your message {certainty} bridged: {msg}",
+            ),
+        )
+
+    async def handle_matrix_leave(self, user: "u.User"):
+        if self.is_direct:
+            self.log.info(
+                f"{user.mxid} left private chat portal with {self.li_other_user_urn}"
+            )
+            if user.li_member_urn == self.li_receiver_urn:
+                self.log.info(
+                    f"{user.mxid} was the recipient of this portal. "
+                    "Cleaning up and deleting..."
+                )
+                await self.cleanup_and_delete()
+        else:
+            self.log.debug(f"{user.mxid} left portal to {self.li_other_user_urn}")
+
     async def handle_matrix_message(
         self,
         sender: "u.User",
@@ -679,8 +813,12 @@ class Portal(DBPortal, BasePortal):
     ):
         try:
             await self._handle_matrix_message(sender, message, event_id)
-        except Exception:
+        except Error as e:
+            self.log.exception(f"Failed handling {event_id}: {e.to_json()}")
+            await self._send_bridge_error(e.to_json())
+        except Exception as e:
             self.log.exception(f"Failed handling {event_id}")
+            await self._send_bridge_error(str(e))
 
     async def _handle_matrix_message(
         self,
@@ -702,12 +840,7 @@ class Portal(DBPortal, BasePortal):
                 sender,
                 cast(TextMessageEventContent, message),
             )
-        elif message.msgtype in (
-            MessageType.AUDIO,
-            MessageType.FILE,
-            MessageType.IMAGE,
-            MessageType.VIDEO,
-        ):
+        elif message.msgtype.is_media:
             await self._handle_matrix_media(
                 event_id,
                 sender,
@@ -715,6 +848,10 @@ class Portal(DBPortal, BasePortal):
             )
         else:
             self.log.warning(f"Unsupported msgtype {message.msgtype} in {event_id}")
+            await self._send_bridge_error(
+                f"messages of {message.msgtype} are not supported.",
+                certain_failure=True,
+            )
             return
 
     async def _send_linkedin_message(
@@ -730,6 +867,9 @@ class Portal(DBPortal, BasePortal):
 
         async with self.require_send_lock(sender.li_member_urn):
             resp = await sender.client.send_message(self.li_thread_urn, message_create)
+            if not resp.value or not resp.value.event_urn:
+                raise Exception("Response value was None.")
+
             message = DBMessage(
                 mxid=event_id,
                 mx_room=self.mxid,
@@ -787,6 +927,7 @@ class Portal(DBPortal, BasePortal):
         attachment = await sender.client.upload_media(
             data, message.body, message.info.mimetype
         )
+        attachment.media_type = attachment.media_type or ""
         await self._send_linkedin_message(
             event_id,
             sender,
@@ -799,7 +940,7 @@ class Portal(DBPortal, BasePortal):
         event_id: EventID,
         redaction_event_id: EventID,
     ):
-        if not self.mxid:
+        if not self.mxid or not sender.client:
             return
 
         message = await DBMessage.get_by_mxid(event_id, self.mxid)
@@ -831,8 +972,8 @@ class Portal(DBPortal, BasePortal):
         reacting_to: EventID,
         reaction: str,
     ):
-        assert sender.li_member_urn
-        assert self.mxid
+        if not sender.li_member_urn or not self.mxid or not sender.client:
+            return
         async with self.require_send_lock(sender.li_member_urn):
             message = await DBMessage.get_by_mxid(reacting_to, self.mxid)
             if not message:
@@ -890,7 +1031,19 @@ class Portal(DBPortal, BasePortal):
         message: ConversationEvent,
     ):
         try:
-            if message.event_content.message_event.recalled_at:
+            if message.subtype == "CONVERSATION_UPDATE":
+                if (
+                    (ec := message.event_content)
+                    and (me := ec.message_event)
+                    and (cc := me.custom_content)
+                    and (nu := cc.conversation_name_update_content)
+                ):
+                    await self._update_name(nu.new_name)
+            elif (
+                (ec := message.event_content)
+                and (me := ec.message_event)
+                and me.recalled_at
+            ):
                 await self._handle_linkedin_message_deletion(sender, message)
             else:
                 await self._handle_linkedin_message(source, sender, message)
@@ -907,11 +1060,12 @@ class Portal(DBPortal, BasePortal):
     ):
         assert self.mxid
         assert self.li_receiver_urn
+        assert message.entity_urn
         li_message_urn = message.entity_urn
 
         # Check in-memory queue for duplicates
         message_exists = False
-        event_ids: List[EventID] = []
+        event_ids: list[EventID] = []
         async with self.require_send_lock(sender.li_member_urn):
             if li_message_urn in self._dedup:
                 self.log.trace(
@@ -960,86 +1114,81 @@ class Portal(DBPortal, BasePortal):
                 await intent.ensure_joined(self.mxid)
                 self._backfill_leave.add(intent)
 
-            message_event = message.event_content.message_event
-            timestamp = message.created_at
+            if (ec := message.event_content) and (message_event := ec.message_event):
+                timestamp = message.created_at or datetime.now()
 
-            event_ids = []
+                event_ids = []
 
-            # Handle subject
-            if message_event.subject:
-                event_ids.append(
-                    await self._send_message(
-                        intent,
-                        linkedin_subject_to_matrix(message_event.subject),
-                        timestamp=timestamp,
-                    )
-                )
-
-            # Handle attachments
-            event_ids.extend(
-                await self._handle_linkedin_attachments(
-                    source, intent, timestamp, message_event.attachments
-                )
-            )
-
-            # Handle custom content
-            if message_event.custom_content:
-                if message_event.custom_content.third_party_media:
-                    event_ids.extend(
-                        await self._handle_linkedin_third_party_media(
-                            source,
-                            intent,
-                            timestamp,
-                            message_event.custom_content.third_party_media,
-                        )
-                    )
-
-                # Handle InMail message text
-                if message_event.custom_content.sp_inmail_content:
+                # Handle subject
+                if message_event.subject:
                     event_ids.append(
                         await self._send_message(
                             intent,
-                            await linkedin_spinmail_to_matrix(
-                                message_event.custom_content.sp_inmail_content
-                            ),
+                            linkedin_subject_to_matrix(message_event.subject),
                             timestamp=timestamp,
                         )
                     )
 
-            # Handle the normal message text itself
-            if message_event.attributed_body and message_event.attributed_body.text:
-                content = await linkedin_to_matrix(message_event.attributed_body)
-                event_ids.append(
-                    await self._send_message(intent, content, timestamp=timestamp)
+                # Handle attachments
+                event_ids.extend(
+                    await self._handle_linkedin_attachments(
+                        source, intent, timestamp, message_event.attachments
+                    )
                 )
-                # TODO (#55) error handling
 
-            event_ids = [event_id for event_id in event_ids if event_id]
-            if len(event_ids) == 0:
-                return
+                # Handle custom content
+                if cc := message_event.custom_content:
+                    if cc.third_party_media:
+                        event_ids.extend(
+                            await self._handle_linkedin_third_party_media(
+                                source,
+                                intent,
+                                timestamp,
+                                cc.third_party_media,
+                            )
+                        )
 
-            # Save all of the messages in the database.
-            self.log.debug(f"Handled LinkedIn message {li_message_urn} -> {event_ids}")
-            await DBMessage.bulk_create(
-                li_message_urn=li_message_urn,
-                li_thread_urn=self.li_thread_urn,
-                li_sender_urn=sender.li_member_urn,
-                li_receiver_urn=self.li_receiver_urn,
-                mx_room=self.mxid,
-                timestamp=timestamp,
-                event_ids=event_ids,
-            )
-            await self._send_delivery_receipt(event_ids[-1])
+                    # Handle InMail message text
+                    if cc.sp_inmail_content:
+                        event_ids.append(
+                            await self._send_message(
+                                intent,
+                                await linkedin_spinmail_to_matrix(cc.sp_inmail_content),
+                                timestamp=timestamp,
+                            )
+                        )
+
+                # Handle the normal message text itself
+                if message_event.attributed_body and message_event.attributed_body.text:
+                    content = await linkedin_to_matrix(message_event.attributed_body)
+                    event_ids.append(
+                        await self._send_message(intent, content, timestamp=timestamp)
+                    )
+                    # TODO (#55) error handling
+
+                event_ids = [event_id for event_id in event_ids if event_id]
+                if len(event_ids) == 0:
+                    return
+
+                # Save all of the messages in the database.
+                self.log.debug(
+                    f"Handled LinkedIn message {li_message_urn} -> {event_ids}"
+                )
+                await DBMessage.bulk_create(
+                    li_message_urn=li_message_urn,
+                    li_thread_urn=self.li_thread_urn,
+                    li_sender_urn=sender.li_member_urn,
+                    li_receiver_urn=self.li_receiver_urn,
+                    mx_room=self.mxid,
+                    timestamp=timestamp,
+                    event_ids=event_ids,
+                )
+                await self._send_delivery_receipt(event_ids[-1])
         # end if message_exists
 
         # Handle reactions
-        reaction_summaries = sorted(
-            message.reaction_summaries,
-            key=lambda r: r.first_reacted_at,
-            reverse=True,
-        )
         reaction_event_id = event_ids[-1]  # react to the last event
-        for reaction_summary in reaction_summaries:
+        for reaction_summary in message.reaction_summaries:
             await self._handle_reaction_summary(
                 li_message_urn,
                 source,
@@ -1077,7 +1226,7 @@ class Portal(DBPortal, BasePortal):
         source: "u.User",
         reaction_event_id: EventID,
         reaction_summary: ReactionSummary,
-    ) -> List[EventID]:
+    ) -> list[EventID]:
         if not reaction_summary.emoji or not source.client:
             return []
 
@@ -1092,7 +1241,6 @@ class Portal(DBPortal, BasePortal):
             sender = await p.Puppet.get_by_li_member_urn(reactor.reactor_urn)
             intent = sender.intent_for(self)
 
-            # TODO (#33) figure out how many reactions should be added
             mxid = await intent.react(
                 self.mxid, reaction_event_id, reaction_summary.emoji
             )
@@ -1100,7 +1248,7 @@ class Portal(DBPortal, BasePortal):
 
             self.log.debug(
                 f"{sender.mxid} reacted to {reaction_event_id} with "
-                "{reaction_summary.emoji}, got {mxid}."
+                f"{reaction_summary.emoji}, got {mxid}."
             )
 
             await DBReaction(
@@ -1119,10 +1267,12 @@ class Portal(DBPortal, BasePortal):
         source: "u.User",
         intent: IntentAPI,
         timestamp: datetime,
-        attachments: List[MessageAttachment],
-    ) -> List[EventID]:
+        attachments: list[MessageAttachment],
+    ) -> list[EventID]:
         event_ids = []
         for attachment in attachments:
+            if not attachment.reference:
+                continue
             url = attachment.reference.string
             if not url:
                 continue
@@ -1141,6 +1291,7 @@ class Portal(DBPortal, BasePortal):
                 file=decryption_info,
                 info=info,
                 msgtype=msgtype,
+                body=attachment.name,
             )
 
             event_id = await self._send_message(
@@ -1160,35 +1311,36 @@ class Portal(DBPortal, BasePortal):
         intent: IntentAPI,
         timestamp: datetime,
         third_party_media: ThirdPartyMedia,
-    ) -> List[EventID]:
+    ) -> list[EventID]:
         if not third_party_media:
             return []
 
         if third_party_media.media_type == "TENOR_GIF":
-            if third_party_media.media.gif.url:
-                msgtype = MessageType.IMAGE
-                mxc, info, decryption_info = await self._reupload_linkedin_file(
-                    third_party_media.media.gif.url,
-                    source,
-                    intent,
-                    encrypt=self.encrypted,
-                    width=third_party_media.media.gif.original_width,
-                    height=third_party_media.media.gif.original_height,
-                )
-                content = MediaMessageEventContent(
-                    url=mxc,
-                    file=decryption_info,
-                    info=info,
-                    msgtype=msgtype,
-                )
+            if not third_party_media.media or not third_party_media.media.gif:
+                return []
+            msgtype = MessageType.IMAGE
+            mxc, info, decryption_info = await self._reupload_linkedin_file(
+                third_party_media.media.gif.url,
+                source,
+                intent,
+                encrypt=self.encrypted,
+                width=third_party_media.media.gif.original_width,
+                height=third_party_media.media.gif.original_height,
+            )
+            content = MediaMessageEventContent(
+                url=mxc,
+                file=decryption_info,
+                info=info,
+                msgtype=msgtype,
+            )
 
-                event_id = await self._send_message(
-                    intent,
-                    content,
-                    timestamp=timestamp,
-                )
-                # TODO (#55) error handling
-                return [event_id]
+            event_id = await self._send_message(
+                intent,
+                content,
+                timestamp=timestamp,
+            )
+            # TODO (#55) error handling
+            return [event_id]
 
         self.log.warning(f"Unsupported third party media: {third_party_media}.")
         return []
@@ -1205,7 +1357,7 @@ class Portal(DBPortal, BasePortal):
         find_size: bool = False,
         width: Optional[int] = None,
         height: Optional[int] = None,
-    ) -> Tuple[ContentURI, MediaInfo, Optional[EncryptedFile]]:
+    ) -> tuple[ContentURI, MediaInfo, Optional[EncryptedFile]]:
         if not url:
             raise ValueError("URL not provided")
 
@@ -1249,8 +1401,12 @@ class Portal(DBPortal, BasePortal):
         sender: "p.Puppet",
         event: RealTimeEventStreamEvent,
     ):
-        assert event.event_urn
-        assert self.li_receiver_urn
+        if (
+            not event.event_urn
+            or not self.li_receiver_urn
+            or not event.reaction_summary
+        ):
+            return
         reaction = event.reaction_summary.emoji
         # Make up a URN for the reacton for dedup purposes
         dedup_id = URN(
