@@ -1,12 +1,13 @@
 import asyncio
+import inspect
 import time
+from asyncio.futures import Future
+from datetime import datetime
 from typing import (
     AsyncGenerator,
     AsyncIterable,
     Awaitable,
     cast,
-    Dict,
-    List,
     Optional,
     TYPE_CHECKING,
 )
@@ -18,7 +19,7 @@ from linkedin_messaging.api_objects import (
     ReactionSummary,
     RealTimeEventStreamEvent,
 )
-from mautrix.bridge import async_getter_lock, BaseUser
+from mautrix.bridge import async_getter_lock, BaseUser, BridgeState
 from mautrix.errors import MNotFound
 from mautrix.types import (
     PushActionType,
@@ -37,6 +38,7 @@ from .db import User as DBUser
 if TYPE_CHECKING:
     from .__main__ import LinkedInBridge
 
+METRIC_CONNECTED = Gauge("bridge_connected", "Bridge users connected to LinkedIn")
 METRIC_LOGGED_IN = Gauge("bridge_logged_in", "Users logged into the bridge")
 METRIC_SYNC_THREADS = Summary("bridge_sync_threads", "calls to sync_threads")
 
@@ -45,8 +47,8 @@ class User(DBUser, BaseUser):
     shutdown: bool = False
     config: Config
 
-    by_mxid: Dict[UserID, "User"] = {}
-    by_li_member_urn: Dict[URN, "User"] = {}
+    by_mxid: dict[UserID, "User"] = {}
+    by_li_member_urn: dict[URN, "User"] = {}
 
     listen_task: Optional[asyncio.Task]
 
@@ -115,7 +117,7 @@ class User(DBUser, BaseUser):
 
     # region Database getters
 
-    def _add_to_cache(self) -> None:
+    def _add_to_cache(self):
         self.by_mxid[self.mxid] = self
         if self.li_member_urn:
             self.by_li_member_urn[self.li_member_urn] = self
@@ -123,7 +125,7 @@ class User(DBUser, BaseUser):
     @classmethod
     async def all_logged_in(cls) -> AsyncGenerator["User", None]:
         users = await super().all_logged_in()
-        for user in cast(List["User"], users):
+        for user in cast(list["User"], users):
             try:
                 yield cls.by_mxid[user.mxid]
             except KeyError:
@@ -188,10 +190,14 @@ class User(DBUser, BaseUser):
         if not self.client or not await self.client.logged_in():
             return False
 
+        if (
+            mp := (await self.client.get_user_profile()).mini_profile
+        ) and mp.entity_urn:
+            self.li_member_urn = mp.entity_urn
+        else:
+            return False
+
         self.log.info("Loaded session successfully")
-        self.li_member_urn = (
-            await self.client.get_user_profile()
-        ).mini_profile.entity_urn
         self._track_metric(METRIC_LOGGED_IN, True)
         self._is_logged_in = True
         self.is_connected = None
@@ -199,7 +205,7 @@ class User(DBUser, BaseUser):
         asyncio.create_task(self.post_login())
         return True
 
-    async def reconnect(self) -> None:
+    async def reconnect(self):
         assert self.listen_task
         self._is_refreshing = True
         await self.listen_task
@@ -220,9 +226,12 @@ class User(DBUser, BaseUser):
 
     async def on_logged_in(self, client: LinkedInMessaging):
         self.client = client
-        self.li_member_urn = (
-            await self.client.get_user_profile()
-        ).mini_profile.entity_urn
+        if (
+            mp := (await self.client.get_user_profile()).mini_profile
+        ) and mp.entity_urn:
+            self.li_member_urn = mp.entity_urn
+        else:
+            raise Exception("No mini_profile.entity_urn on the user profile!")
         await self.save()
         self.stop_listen()
         asyncio.create_task(self.post_login())
@@ -243,10 +252,14 @@ class User(DBUser, BaseUser):
         self.start_listen()
 
     async def logout(self):
+        self.log.info("Logging out")
         if self.listen_task:
+            self.log.info("Cancelling the listen task.")
             self.listen_task.cancel()
         if self.client:
+            self.log.info("Logging out the client.")
             await self.client.logout()
+        await self.push_bridge_state(ok=False, error="logged-out")
         puppet = await pu.Puppet.get_by_li_member_urn(self.li_member_urn, create=False)
         if puppet and puppet.is_real_user:
             await puppet.switch_mxid(None, None)
@@ -266,12 +279,12 @@ class User(DBUser, BaseUser):
 
     # region Thread Syncing
 
-    async def get_direct_chats(self) -> Dict[UserID, List[RoomID]]:
+    async def get_direct_chats(self) -> dict[UserID, list[RoomID]]:
         assert self.li_member_urn
         return {
             pu.Puppet.get_mxid_from_id(portal.li_other_user_urn): [portal.mxid]
             async for portal in po.Portal.get_all_by_li_receiver_urn(self.li_member_urn)
-            if portal.mxid
+            if portal.mxid and portal.li_other_user_urn
         }
 
     @async_time(METRIC_SYNC_THREADS)
@@ -287,20 +300,44 @@ class User(DBUser, BaseUser):
         except Exception:
             self.log.exception("Failed to sync threads")
 
-    async def _sync_threads(self) -> None:
+    async def _sync_threads(self):
         assert self.client
         sync_count = self.config["bridge.initial_chat_sync"]
         if sync_count <= 0:
             return
 
         self.log.debug("Fetching threads...")
-        # user_portals = await UserPortal.all(self.li_member_urn)
 
-        async for conversation in self.client.get_all_conversations():
-            try:
-                await self._sync_thread(conversation)
-            except Exception:
-                self.log.exception(f"Failed to sync thread {conversation.entity_urn}")
+        last_activity_before = datetime.now()
+        synced_threads = 0
+        while True:
+            if synced_threads >= sync_count:
+                break
+            conversations_response = await self.client.get_conversations(
+                last_activity_before=last_activity_before
+            )
+            for conversation in conversations_response.elements:
+                if synced_threads >= sync_count:
+                    break
+                try:
+                    await self._sync_thread(conversation)
+                except Exception:
+                    self.log.exception(
+                        f"Failed to sync thread {conversation.entity_urn}"
+                    )
+                synced_threads += 1
+
+            await self.update_direct_chats()
+
+            # The page size is 20, by default, so if we get less than 20, we are at the
+            # end of the list so we should stop.
+            if len(conversations_response.elements) < 20:
+                break
+
+            if last_activity_at := conversations_response.elements[-1].last_activity_at:
+                last_activity_before = last_activity_at
+            else:
+                break
 
         await self.update_direct_chats()
 
@@ -310,7 +347,16 @@ class User(DBUser, BaseUser):
         li_other_user_urn = None
         if not conversation.group_chat:
             other_user = conversation.participants[0]
-            li_other_user_urn = other_user.messaging_member.mini_profile.entity_urn
+            if (
+                (mm := other_user.messaging_member)
+                and (mp := mm.mini_profile)
+                and mp.entity_urn
+            ):
+                li_other_user_urn = mp.entity_urn
+                if li_other_user_urn == URN("UNKNOWN"):
+                    li_other_user_urn = conversation.entity_urn
+            else:
+                raise Exception("Other chat participant didn't have an entity_urn!")
 
         portal = await po.Portal.get_by_li_thread_urn(
             conversation.entity_urn,
@@ -354,30 +400,100 @@ class User(DBUser, BaseUser):
 
     # endregion
 
-    # region Listener Management
+    # region Listener and State Management
+
+    async def fill_bridge_state(self, state: BridgeState):
+        await super().fill_bridge_state(state)
+        state.remote_id = self.li_member_urn.get_id()
+        state.remote_name = ""
+        user = await User.get_by_li_member_urn(self.li_member_urn)
+        if user and user.client:
+            try:
+                if mp := (await user.client.get_user_profile()).mini_profile:
+                    state.remote_name = " ".join(
+                        n for n in [mp.first_name, mp.last_name] if n
+                    )
+            except Exception:
+                pass
+
+    async def get_bridge_state(self) -> BridgeState:
+        if not self.client or not await self.client.logged_in():
+            return BridgeState(ok=False, error="logged-out")
+        elif not self.listen_task or self.listen_task.done() or not self.is_connected:
+            return BridgeState(ok=False, error="li-no-listener")
+        return BridgeState(ok=True)
 
     def stop_listen(self):
+        self.log.info("Stopping the listener.")
+        curframe = inspect.currentframe()
+        callstack = "\n".join([str(f) for f in inspect.getouterframes(curframe, 10)])
+        self.log.info(f"Call stack:\n{callstack}")
         if self.listen_task:
+            self.log.info("Cancelling the listen task.")
             self.listen_task.cancel()
         self.listen_task = None
 
+    def on_listen_task_end(self, future: Future):
+        if future.cancelled():
+            self.log.info("Listener task cancelled")
+        if not self.shutdown:
+            self.start_listen()
+
+    listener_event_handlers_created: bool = False
+    listener_task_i: int = 0
+
     def start_listen(self):
-        self.listen_task = asyncio.create_task(self._try_listen())
+        self.log.info("Starting listener task.")
+        self.listen_task = asyncio.create_task(
+            self._try_listen(),
+            name=f"listener task #{self.listener_task_i}",
+        )
+        self.listen_task.add_done_callback(self.on_listen_task_end)
 
     async def _try_listen(self):
         assert self.client
-        self.client.add_event_listener("event", self.handle_linkedin_event)
-        self.client.add_event_listener(
-            "reactionAdded", self.handle_linkedin_reaction_added
-        )
+        if not self.listener_event_handlers_created:
+            self.client.add_event_listener(
+                "ALL_EVENTS", self.handle_linkedin_stream_event
+            )
+            self.client.add_event_listener(
+                "STREAM_ERROR", self.handle_linkedin_listener_error
+            )
+            self.client.add_event_listener("event", self.handle_linkedin_event)
+            self.client.add_event_listener(
+                "reactionAdded", self.handle_linkedin_reaction_added
+            )
+            self.listener_event_handlers_created = True
         await self.client.start_listener()
 
+    async def handle_linkedin_stream_event(self, _):
+        self._track_metric(METRIC_CONNECTED, True)
+        await self.push_bridge_state(ok=True)
+
+    async def handle_linkedin_listener_error(self, error: Exception):
+        self._track_metric(METRIC_CONNECTED, False)
+        await self.push_bridge_state(
+            ok=False,
+            error="li-connection-error",
+            message=str(error),
+        )
+
     async def handle_linkedin_event(self, event: RealTimeEventStreamEvent):
         assert self.client
         assert isinstance(event.event, ConversationEvent)
+        assert event.event.entity_urn
 
         thread_urn, message_urn = map(URN, event.event.entity_urn.id_parts)
-        sender_urn = event.event.from_.messaging_member.mini_profile.entity_urn
+        if (
+            (e := event.event)
+            and (f := e.from_)
+            and (mm := f.messaging_member)
+            and (mp := mm.mini_profile)
+            and (entity_urn := mp.entity_urn)
+        ):
+            sender_urn = entity_urn
+        else:
+            raise Exception("Invalid sender: no entity_urn found!", event)
 
         portal = await po.Portal.get_by_li_thread_urn(
             thread_urn,
@@ -385,11 +501,11 @@ class User(DBUser, BaseUser):
             create=False,
         )
         if not portal:
-            # Force a thread sync for all of the recent conversations. This should be a
-            # noop for most of them except the newly created conversation.
             conversations = await self.client.get_conversations()
             for conversation in conversations.elements:
-                await self._sync_thread(conversation)
+                if conversation.entity_urn == thread_urn:
+                    await self._sync_thread(conversation)
+                    break
 
             # Nothing more to do, since the backfill should handle the message coming
             # in.
